#!/usr/bin/perl
#$rcs = ' $Id: indexer.pl,v 1.21 1999/07/26 22:37:32 perlfect Exp $ ' ;	

use Fcntl;
use locale;

package AnyDBM_File;
@ISA = qw(DB_File GDBM_File SDBM_File ODBM_File NDBM_File) unless @ISA;

for (@ISA) {
  last if eval "require $_";
}

package main;
my $VERSION = "3.01";
require './conf.pl';

$|=1;

my $DN  = 0;  #documents number
my $TN  = 0;  #terms number

my @stopwords;
my @no_index;

my %inv_index_db;
my %docs_db;
my %desc_db;
my %title_db;
my %terms_db;
my %df_db;
my %tf_db;

cleanup();

tie %inv_index_db, AnyDBM_File, $INV_INDEX_DB_FILE, O_CREAT|O_RDWR, 0755 or die "Cannot open $INV_INDEX_DB_FILE: $!";   
tie %docs_db,      AnyDBM_File, $DOCS_DB_FILE, O_CREAT|O_RDWR, 0755      or die "Cannot open $DOCS_DB_FILE: $!";   
tie %desc_db,      AnyDBM_File, $DESC_DB_FILE, O_CREAT|O_RDWR, 0755      or die "Cannot open $DESC_DB_FILE: $!";   
tie %titles_db,    AnyDBM_File, $TITLES_DB_FILE, O_CREAT|O_RDWR, 0755    or die "Cannot open $TITLES_DB_FILE: $!";   
tie %terms_db,     AnyDBM_File, $TERMS_DB_FILE, O_CREAT|O_RDWR, 0755     or die "Cannot open $TERMS_DB_FILE: $!";   
tie %df_db,        AnyDBM_File, $DF_DB_FILE, O_CREAT|O_RDWR, 0755        or die "Cannot open $DF_DB_FILE: $!";
tie %tf_db,        AnyDBM_File, $TF_DB_FILE, O_CREAT|O_RDWR, 0755        or die "Cannot open $TF_DB_FILE: $!";

print "Loading \'no index\' regular expressions:\n";
load_excludes();
print "Loading stopwords...";
load_stopwords();
print "Done.\n";

print "Starting crawler...\n";
crawl($DOCUMENT_ROOT.$INDEXER_START_DIR);
print "Crawler finished($DN files, $TN terms)\n\n";

print "Calculating weight vectors: \n";
weights();

print "Removing unused db files:";

print "$TF_DB_FILE...";
if (unlink $TF_DB_FILE) {
  print "ok\n";
} else {
  warn "Cannot unlink $TF_DB_FILE: $!";
}

print "$DF_DB_FILE...";
if (unlink $DF_DB_FILE) {
  print "ok\n";
} else {
  warn "Cannot unlink $DF_DB_FILE: $!";
}

print "Indexer finished.\n";

sub crawl {
  my $dir = $_[0];
  my $doc_id;
  my $file;
  
  print $dir,"\n";

  chdir $dir or (warn "Cannot chdir $dir: $!" and next);
  opendir(DIR, $dir) or (warn "Cannot open $dir: $!" and next);
  my @contents = readdir DIR;
  closedir(DIR);

  my @dirs  = grep {-d and not /^\.{1,2}$/} @contents; 
  my @files = grep {-f and /^.+\.(.+)$/ and grep {/^$1$/} @EXT} @contents;
  
  FILE: foreach my $f (@files) {
    $file = $dir."/".$f;
    $file =~ s/\/\//\//og;

    foreach my $regexp (@no_index) {
      next FILE if $file =~ m/^$regexp$/;
    }
    
    $doc_id = record_file($file);
    index_file($file, $doc_id);
  }

  DIR: foreach my $d (@dirs) {
    $file = $dir."/".$d;
    $file =~ s/\/\//\//og;
    
    foreach my $regexp (@no_index) {
      next DIR if $file =~ /^$regexp$/;
    }
    crawl($file);
  }
}

sub index_file {
  my ($file, $doc_id) = @_;
  my ($term, $term_id);
  my $buffer;
  my %tf;

  undef $/;
  open(FILE, $file) or (warn "Cannot open $file: $!" and next);
  $buffer = <FILE>;
  close(FILE); 
  $/ = "\n";  

  print "\t $doc_id: $file\n";

  record_desc($doc_id, $buffer);    
  $buffer = normalize($buffer);
  
  foreach (split " ", $buffer) {
    if ($term = stem($_)) {
      $term_id = record_term($term);
      ++$tf{$term_id};
    }
  }
  
  foreach (keys %tf) {
    $df_db{$_}++;
    $tf_db{$_} = '' unless defined $tf_db{$_};
    $tf_db{$_} .= pack("ww", $doc_id, $tf{$_}); 
  }
}

sub weights {
  my ($weight, $term_id, $doc_id);
  my $step = $TN / 50;
  my $count = 0;
  
  print "0%  10%  20%  30%  40%  50%  60%  70%  80%  90%  100%\n";
  print "|----|----|----|----|----|----|----|----|----|----|\n";

  foreach $term_id (keys %tf_db) {
    my %tdf = unpack("w*",$tf_db{$term_id}); 
    $count++;
    print "\b >" unless $count % $step;
    foreach $doc_id (keys %tdf) {
      #print "weight = $tdf{$doc_id} * log ($DN / $df_db{$term_id})\n";
      $weight = $tdf{$doc_id} * log ($DN / $df_db{$term_id});
      $inv_index_db{$term_id} = '' unless $inv_index_db{$term_id};
      $inv_index_db{$term_id} .= pack("ff", $doc_id, $weight);
    }
  }
  print "\n";
}


sub normalize {
  my $buffer = $_[0];
  my $stopword;

  $buffer =~ s/<.*?>/ /gs;
  $buffer =~ s/-//gs;
  $buffer =~ s/\W/ /gs;
  $buffer =~ s/\b\d+\b/ /gs;

  foreach (@stopwords) {
    $buffer =~ s/\b$_\b//gis;
  }

  $buffer =~ s/[\n\s]+/ /gs;
  $buffer = lc $buffer;

  return $buffer;
}

sub record_file {
  my $file = $_[0];

  $file =~ m/^$DOCUMENT_ROOT(.*)$/;
  $file = $1;
  unless ($file =~ /^\//) {
    $file = "/".$file;
  }

  ++$DN;
  $docs_db{$DN} = $file;
  return $DN;
}

sub record_desc {
  my ($doc_id, $buffer) = @_;
  my ($desc, $title);
  my @desc_ary;

  $buffer =~ m/<META\s+name\s*=\s*[\"\']?description[\"\']? content=[\"\']?(.*?)[\"\']?>/is;
  $desc = $1;
  
  unless ($desc) {
    $buffer =~ m/<BODY.*?>(.*)<\/BODY>/is;
    $desc = $1;
    $desc =~ s/<.+?>/ /gis; #strip html
    $desc =~ s/[\n\r]/ /gis;
    @desc_ary = split " ", $desc;
    $desc = $1;
    $desc = join " ", @desc_ary[0..$DESC_WORDS];
  }
  $desc_db{$doc_id} = $desc;

  $buffer =~ m/<TITLE>(.*?)<\/TITLE>/is;
  $title = $1;
  $title ||= "Untitled";
  $titles_db{$doc_id} = $title;
}

sub record_term {
  my $term = $_[0];

  print STDERR "Warning: record_term($term): No term was supplied\n" unless $term;

  if ($terms_db{$term}) {
    return $terms_db{$term};
  } else {
    ++$TN;
    $terms_db{$term} = $TN;
    return $TN;
  }
}

sub load_excludes {
  if (-e $NO_INDEX_FILE) {
    open (FILE, $NO_INDEX_FILE) or (warn "Cannot open $NO_INDEX_FILE: $!" and next);
    while (<FILE>) {
      chomp;
      $_ =~ s/(\#.*)//g;    #ingore comments
      $_ =~ s/[\/\s]*$//;   #remove any trailing spaces and slashes
      print "\t- $_\n";
      
      $_ = quotemeta;       #escape all non-alphanumeric characters
      $_ =~ s/\\\*/\.\*/g;  #except for the * which is replaced by .*
      push @no_index, $_;
    }
    close (FILE);
  } else {
    print STDERR "Warning: $NO_INDEX_FILE missing.";
  }
}

sub load_stopwords {
  open (FILE, $STOPWORDS_FILE) or (warn "Cannot open $STOPWORDS_FILE: $!" and next);
  while (<FILE>) {
    chomp;
    push @stopwords, $_;
  }
}

sub stem {
  my $str = $_[0];
  #$str =~ tr/A-Za-z0-9//cs;        #delete any non-alphanumeric characters
  if (length $str >= $MINLENGTH) { #check the length of the resulting string
    return substr $str, 0, $STEMCHARS;
  } else {
    return 0;
  }
}

sub cleanup {
  print "Deleting existing db files:\n";
  foreach (($TERMS_DB_FILE, $INV_INDEX_DB_FILE, $DF_DB_FILE, $DOCS_DB_FILE, $DESC_DB_FILE, $TITLES_DB_FILE, $TF_DB_FILE)) {
    if (-e $_) {
      print "\t $_\n";
      unlink $_ or (warn "Cannot unlink $_: $!" and next);
    }
  }
}
